"use strict";
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.toStringSingle = exports.toStringMulti = exports.or = exports.and = exports.whereIn = exports.where = exports.search = exports.offset = exports.limit = exports.sort = exports.exclude = exports.fields = exports.query = void 0;
const types_1 = require("./types");
function query(queryEndpoint, queryName) {
    return (builder => {
        return Object.assign(Object.assign({}, builder), { queryEndpoint,
            queryName });
    });
}
exports.query = query;
/**
 * Select only given fields.
 * Resulting response type is narrowed to only include specified fields (including nested ones).
 *
 * @example
 * ```ts
 * request().pipe(
 *   fields('*'),                   // select all fields
 *   fields(["name", "genres.id"]), // select specific fields, incuding sub-type
 *   fields(["name", "genres.*"]),  // select specific fields, incuding all fields of sub-type
 * )
 * ```
 * @see {@link https://api-docs.igdb.com/?shell#fields}
 * @param f
 */
function fields(f) {
    if (Array.isArray(f)) {
        const fieldsString = f.join(",").replace(/\s/g, '');
        return (builder => {
            return Object.assign(Object.assign({}, builder), { queryFields: Object.assign(Object.assign({}, builder.queryFields), { fields: `fields ${fieldsString}` }) });
        });
    }
    return (builder => {
        return Object.assign(Object.assign({}, builder), { queryFields: Object.assign(Object.assign({}, builder.queryFields), { fields: `fields *` }) });
    });
}
exports.fields = fields;
/**
 * Exclude given fields from selection.
 * Resulting response type is narrowed to exclude specified fields (not compatible with nested fields).
 *
 * @example
 * ```ts
 * request().pipe(
 *   fields('*'),
 *   exclude(["name", "created_at"]), // exclude specific fields
 * )
 * ```
 * @see {@link https://api-docs.igdb.com/?shell#exclude}
 * @param exclude
 */
function exclude(exclude) {
    const fieldsString = exclude.join(",").replace(/\s/g, '');
    return (builder => {
        return Object.assign(Object.assign({}, builder), { queryFields: Object.assign(Object.assign({}, builder.queryFields), { exclude: `exclude ${fieldsString}` }) });
    });
}
exports.exclude = exclude;
/**
 * Sort results by specified field.
 *
 * @example
 * ```ts
 * request().pipe(
 *   sort('name'),          // defaults to 'asc' when direction is not specified
 *   sort('name', 'desc'),  // manually specify direction
 * )
 * ```
 * @see {@link https://api-docs.igdb.com/?shell#sorting}
 * @param field
 * @param direction
 */
function sort(field, direction = 'asc') {
    return builder => {
        return Object.assign(Object.assign({}, builder), { queryFields: Object.assign(Object.assign({}, builder.queryFields), { sort: `sort ${String(field)} ${direction}` }) });
    };
}
exports.sort = sort;
/**
 * Limit the number of results returned by the query.
 *
 * @example
 * ```ts
 * request().pipe(
 *   limit(42), // Only retrieve 42 elements
 * )
 * ```
 * @see {@link https://api-docs.igdb.com/?shell#pagination}
 * @param limit
 */
function limit(limit) {
    return builder => {
        return Object.assign(Object.assign({}, builder), { queryFields: Object.assign(Object.assign({}, builder.queryFields), { limit: `limit ${limit}` }) });
    };
}
exports.limit = limit;
/**
 * Start results at a given offset.
 *
 * @example
 * ```ts
 * request().pipe(
 *   offset(50), // Start he results at position 50...
 *   limit(42),  // ...and limit the number of results to 42
 * )
 * ```
 * @see {@link https://api-docs.igdb.com/?shell#pagination}
 * @param offset
 */
function offset(offset) {
    return builder => {
        return Object.assign(Object.assign({}, builder), { queryFields: Object.assign(Object.assign({}, builder.queryFields), { offset: `offset ${offset}` }) });
    };
}
exports.offset = offset;
/**
 * Search based on name, results are sorted by similarity to the given search string.
 *
 * @example
 * ```ts
 * request().pipe(
 *   search('zelda'),
 * )
 * ```
 * @see {@link https://api-docs.igdb.com/?shell#search-1}
 * @param search
 */
function search(search) {
    return builder => {
        return Object.assign(Object.assign({}, builder), { queryFields: Object.assign(Object.assign({}, builder.queryFields), { search: `search "${encodeString(search)}"` }) });
    };
}
exports.search = search;
function encodeWhereParam(value, flag) {
    let v = null;
    if (typeof value === 'boolean' || value === null) {
        return String(value);
    }
    if (typeof flag === 'number') {
        if ((flag & types_1.WhereFlags.STRING) === types_1.WhereFlags.STRING) {
            v = `"${encodeString(value)}"`;
        }
        if ((flag & types_1.WhereFlags.STARTSWITH) === types_1.WhereFlags.STARTSWITH) {
            v += '*';
        }
        if ((flag & types_1.WhereFlags.ENDSWITH) === types_1.WhereFlags.ENDSWITH) {
            v = '*' + v;
        }
        if ((flag & types_1.WhereFlags.RAW) === types_1.WhereFlags.RAW) {
            v = value;
        }
    }
    if (typeof value === 'string' && v === null) {
        v = `"${encodeString(value)}"`;
    }
    if (v === null) {
        v = value;
    }
    return v;
}
function getOp(flag) {
    if ((flag & types_1.WhereInFlags.NAND) === types_1.WhereInFlags.NAND) {
        return '!=';
    }
    if ((flag & types_1.WhereInFlags.NOR) === types_1.WhereInFlags.NOR) {
        return '!=';
    }
    return '=';
}
function encodeWhereInParam(values, flag) {
    const joined = values.map(v => encodeWhereParam(v, flag)).join(',');
    if ((flag & types_1.WhereInFlags.NAND) === types_1.WhereInFlags.NAND) {
        return `[${joined}]`;
    }
    if ((flag & types_1.WhereInFlags.NOR) === types_1.WhereInFlags.NOR) {
        return `(${joined})`;
    }
    if ((flag & types_1.WhereInFlags.AND) === types_1.WhereInFlags.AND) {
        return `[${joined}]`;
    }
    if ((flag & types_1.WhereInFlags.OR) === types_1.WhereInFlags.OR) {
        return `(${joined})`;
    }
    if ((flag & types_1.WhereInFlags.EXACT) === types_1.WhereInFlags.EXACT) {
        return `{${joined}}`;
    }
    throw new Error('WhereInFlags not specified or invalid');
}
/**
 * Filters results based on given parameters. All parameters are type validated if `request` is given a type.
 * Also see {@link whereIn}.
 *
 * @example
 * ```ts
 * request<MyType>().pipe(
 *   where('ratings', '>', 4),  // ratings greater than 4
 *   where('ratings', '>=', 4), // ratings greater than or equal 4
 *   where('ratings', '<', 4),  // ratings less than 4
 *   where('ratings', '<=', 4), // ratings less than or equal 4
 *   where('name', '=', 'zelda'),   // name is zelda (case sensitive)
 *   where('name', '~', 'zelda'),   // name is zelda (case insensitive)
 *   where('name', '!=', 'zelda'),  // name is not zelda
 *   where('name', '=', 'zelda', WhereFlags.STARTSWITH),  // name starts with zelda (also works with ~)
 *   where('name', '=', 'zelda', WhereFlags.ENDSWITH),    // name ends with zelda (also works with ~)
 *   where('name', '=', 'zelda', WhereFlags.CONTAINS),    // name contains zelda (also works with ~)
 *   where('release_date', '=', null)   // no release date
 *   where('release_date', '!=', null)  // release date has any value
 * )
 * ```
 * @see {@link https://api-docs.igdb.com/?shell#filters}
 * @param key
 * @param op
 * @param value
 * @param flag
 */
function where(key, op, value, flag) {
    return builder => {
        return Object.assign(Object.assign({}, builder), { queryFields: Object.assign(Object.assign({}, builder.queryFields), { where: [`${key} ${op} ${encodeWhereParam(value, flag)}`] }) });
    };
}
exports.where = where;
/**
 * Filters results based on given parameters. All parameters are type validated if `request` is given a type.
 * Also see {@link where}.
 *
 * @example
 * ```ts
 * request<MyType>().pipe(
 *   whereIn('genres', [1, 2], WhereInFlags.AND),   // Results whose genres includes 1 and 2
 *   whereIn('genres', [1, 2], WhereInFlags.OR),    // Results whose genres includes 1 or 2
 *   whereIn('genres', [1, 2], WhereInFlags.NAND),  // Results whose genres does not contain both 1 and 2, but can be 1 or 2
 *   whereIn('genres', [1, 2], WhereInFlags.NOR),   // Results whose genres does not contain 1 or does not contain 2
 *   whereIn('genres', [1, 2], WhereInFlags.EXACT), // Results whose exclusive genres are 1 and 2
 * )
 * ```
 * @see {@link https://api-docs.igdb.com/?shell#filters}
 * @param key
 * @param values
 * @param flag
 */
function whereIn(key, values, flag = types_1.WhereInFlags.OR) {
    return builder => {
        return Object.assign(Object.assign({}, builder), { queryFields: Object.assign(Object.assign({}, builder.queryFields), { where: [`${key} ${getOp(flag)} ${encodeWhereInParam(values, flag)}`] }) });
    };
}
exports.whereIn = whereIn;
function groupWhere(orAnd, ...operators) {
    return builder => {
        const sub = operators.flatMap(op => [...op(builder).queryFields.where, ` ${orAnd} `]);
        if (sub.length === 0)
            return builder;
        sub.pop(); // remove last ' ${orAnd} '
        return Object.assign(Object.assign({}, builder), { queryFields: Object.assign(Object.assign({}, builder.queryFields), { where: ['(', ...sub, ')'] }) });
    };
}
/**
 * Combine multiple {@link where} and {@link whereIn} conditions.
 *
 * @example
 * ```ts
 * // Select results that are on platform 1 and 2, and whose name are either 'zelda' or 'link'
 * request<MyType>().pipe(
 *   and(
 *     whereIn('platforms', [1, 2], WhereInFlags.AND),
 *     or(
 *       where('name', '=', 'zelda'),
 *       where('name', '=', 'link'),
 *     )
 *   )
 * )
 * ```
 * @see {@link https://api-docs.igdb.com/?shell#filters}
 * @param operators
 */
function and(...operators) {
    return groupWhere('&', ...operators);
}
exports.and = and;
/**
 * Combine multiple {@link where} and {@link whereIn} conditions.
 *
 * @example
 * ```ts
 * // Select results that are on platform 1 and 2, and whose name are either 'zelda' or 'link'
 * request<MyType>().pipe(
 *   and(
 *     whereIn('platforms', [1, 2], WhereInFlags.AND),
 *     or(
 *       where('name', '=', 'zelda'),
 *       where('name', '=', 'link'),
 *     )
 *   )
 * )
 * ```
 * @see {@link https://api-docs.igdb.com/?shell#filters}
 * @param operators
 */
function or(...operators) {
    return groupWhere('|', ...operators);
}
exports.or = or;
function toStringMulti(builders) {
    return builders.map(b => `query ${b.queryEndpoint} "${b.queryName}" { ${toStringSingle(b)} };`).join("");
}
exports.toStringMulti = toStringMulti;
function toStringSingle(builder) {
    const _a = builder.queryFields, { where } = _a, rest = __rest(_a, ["where"]);
    const w = where.length > 0 ? "where " + where.join("") + ";" : "";
    const r = Object.keys(rest).length > 0 ? Object.values(rest).join(";") + ";" : "";
    return r + w;
}
exports.toStringSingle = toStringSingle;
function encodeString(v) {
    return v.replace(/\\/g, '\\\\').replace(/"/g, '\\"');
}
//# sourceMappingURL=builder.js.map