import type { Builder, ExecutorMulti, ResultMultiMono, NamedBuilder, Pipe, PipeSub, Stringifiable } from "./types";
export type { Builder, BuilderOperator, BuilderOperatorNarrow, ExecutorMulti, ResultMultiMono, NamedBuilder, Options, Pipe, PipeSub, Stringifiable, Executor, NamedBuilderOperator, } from "./types";
export { WhereFlags, WhereInFlags, } from "./types";
export { and, exclude, fields, limit, offset, or, search, sort, where, whereIn } from "./builder";
/**
 * Prepare a request to an apicalypse endpoint.
 *
 * @example
 * ```ts
 * // MyType is optional but recommended, so that results are properly typed
 * request<MyType>().pipe(
 *   fields(["name"]),
 *   sort("created_at", "asc"),
 *   where("created_at", ">",  now),
 * )
 * .execute('http:...')
 * .then(results => ...);
 * ```
 */
export declare function request<T extends Record<any, any>, mode extends 'result' | 'count' = 'result'>(): {
    pipe: Pipe<T, mode>;
    sub: <S extends string>(queryEndpoint: string, queryName: S) => {
        pipe: PipeSub<T, S, mode>;
    };
};
/**
 * Prepare a multi-query request to an apicalypse endpoint.
 *
 * @example
 * ```ts
 * multi(
 *   // MyType is optional but recommended, so that results are properly typed
 *   request<MyType>()
 *     .sub('endpoint', 'alias') // mandatory for multi requests
 *     .pipe(
 *       fields(["name"]),
 *       sort("created_at", "asc"),
 *       where("created_at", ">",  now),
 *     ),
 *   request<MyType>()
 *     .sub('endpoint', 'alias2') // mandatory for multi requests
 *     .pipe(
 *       fields(["name"]),
 *       sort("created_at", "asc"),
 *       where("created_at", "<",  now),
 *     )
 * )
 * .execute('http:...')
 * .then(response => ...);
 * ```
 * @see {@link https://api-docs.igdb.com/?shell#multi-query}
 * @param builders
 */
export declare function multi<T extends Record<any, any>, B extends Builder<T>[]>(...builders: B): Stringifiable & ExecutorMulti<B>;
/**
 * Narrow multi-query results types.
 *
 * @example
 * ```ts
 * interface MyType {
 *   name: string,
 *   created_at: number
 * }
 *
 * multi(
 *  request<MyType>().sub("games", "latest-games").pipe(
 *    fields(["name"]),
 *    sort("created_at", "desc"),
 *    where("created_at", "<",  now),
 *  ),
 *  request<MyType>().sub("games", "coming-soon").pipe(
 *    fields(["created_at"]),
 *    sort("created_at", "desc"),
 *    where("created_at", ">",  now),
 *  )
 * ).execute('http:...').then(response => {
 *   for (const result of response.data) {
 *     if (isNamed(result, 'latest-games')) {
 *       result.result.name // result is of type `{ name: string }`
 *     } else {
 *       result.result.created_at  // result is of type `{ created_at: number }`
 *     }
 *   }
 * });
 * ```
 * @param builder
 * @param name
 */
export declare function isNamed<T extends NamedBuilder<any, string>, S extends string>(builder: ResultMultiMono<T>, name: S): builder is ResultMultiMono<Extract<T, NamedBuilder<any, S>>>;
