"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.isNamed = exports.multi = exports.request = exports.whereIn = exports.where = exports.sort = exports.search = exports.or = exports.offset = exports.limit = exports.fields = exports.exclude = exports.and = exports.WhereInFlags = exports.WhereFlags = void 0;
const axios_1 = __importDefault(require("axios"));
const builder_1 = require("./builder");
var types_1 = require("./types");
Object.defineProperty(exports, "WhereFlags", { enumerable: true, get: function () { return types_1.WhereFlags; } });
Object.defineProperty(exports, "WhereInFlags", { enumerable: true, get: function () { return types_1.WhereInFlags; } });
var builder_2 = require("./builder");
Object.defineProperty(exports, "and", { enumerable: true, get: function () { return builder_2.and; } });
Object.defineProperty(exports, "exclude", { enumerable: true, get: function () { return builder_2.exclude; } });
Object.defineProperty(exports, "fields", { enumerable: true, get: function () { return builder_2.fields; } });
Object.defineProperty(exports, "limit", { enumerable: true, get: function () { return builder_2.limit; } });
Object.defineProperty(exports, "offset", { enumerable: true, get: function () { return builder_2.offset; } });
Object.defineProperty(exports, "or", { enumerable: true, get: function () { return builder_2.or; } });
Object.defineProperty(exports, "search", { enumerable: true, get: function () { return builder_2.search; } });
Object.defineProperty(exports, "sort", { enumerable: true, get: function () { return builder_2.sort; } });
Object.defineProperty(exports, "where", { enumerable: true, get: function () { return builder_2.where; } });
Object.defineProperty(exports, "whereIn", { enumerable: true, get: function () { return builder_2.whereIn; } });
function apicalypse(builder, options = {}) {
    function getAxiosRequestConfig(url) {
        const opts = Object.assign(Object.assign({ queryMethod: 'body', method: 'post' }, options), { url: url });
        const raw = builder.toApicalypseString();
        if (opts.queryMethod === 'url') {
            opts.params = {
                apicalypse: encodeURIComponent(raw),
            };
        }
        else if (opts.queryMethod === 'body') {
            opts.data = raw;
        }
        return opts;
    }
    return {
        execute(url) {
            return axios_1.default.create()(getAxiosRequestConfig(url));
        }
    };
}
/**
 * Prepare a request to an apicalypse endpoint.
 *
 * @example
 * ```ts
 * // MyType is optional but recommended, so that results are properly typed
 * request<MyType>().pipe(
 *   fields(["name"]),
 *   sort("created_at", "asc"),
 *   where("created_at", ">",  now),
 * )
 * .execute('http:...')
 * .then(results => ...);
 * ```
 */
function request() {
    function _pipe(...steps) {
        return steps.reduce((output, f) => f(output), newBuilder());
    }
    const pipe = (...steps) => {
        const builder = _pipe(...steps);
        return {
            execute(url, options = {}) {
                return apicalypse(builder, options).execute(url);
            },
            toApicalypseString() {
                return (0, builder_1.toStringSingle)(builder);
            },
        };
    };
    const sub = (queryEndpoint, queryName) => {
        const pipe = (...steps) => {
            return _pipe((0, builder_1.query)(queryEndpoint, queryName), ...steps);
        };
        return {
            pipe,
        };
    };
    return {
        pipe,
        sub,
    };
}
exports.request = request;
/**
 * Prepare a multi-query request to an apicalypse endpoint.
 *
 * @example
 * ```ts
 * multi(
 *   // MyType is optional but recommended, so that results are properly typed
 *   request<MyType>()
 *     .sub('endpoint', 'alias') // mandatory for multi requests
 *     .pipe(
 *       fields(["name"]),
 *       sort("created_at", "asc"),
 *       where("created_at", ">",  now),
 *     ),
 *   request<MyType>()
 *     .sub('endpoint', 'alias2') // mandatory for multi requests
 *     .pipe(
 *       fields(["name"]),
 *       sort("created_at", "asc"),
 *       where("created_at", "<",  now),
 *     )
 * )
 * .execute('http:...')
 * .then(response => ...);
 * ```
 * @see {@link https://api-docs.igdb.com/?shell#multi-query}
 * @param builders
 */
function multi(...builders) {
    return {
        toApicalypseString() {
            return (0, builder_1.toStringMulti)(builders);
        },
        execute(url, options = {}) {
            return apicalypse(this, options).execute(url);
        }
    };
}
exports.multi = multi;
function newBuilder() {
    return {
        queryFields: {
            where: []
        },
        toApicalypseString() {
            return (0, builder_1.toStringSingle)(this);
        }
    };
}
/**
 * Narrow multi-query results types.
 *
 * @example
 * ```ts
 * interface MyType {
 *   name: string,
 *   created_at: number
 * }
 *
 * multi(
 *  request<MyType>().sub("games", "latest-games").pipe(
 *    fields(["name"]),
 *    sort("created_at", "desc"),
 *    where("created_at", "<",  now),
 *  ),
 *  request<MyType>().sub("games", "coming-soon").pipe(
 *    fields(["created_at"]),
 *    sort("created_at", "desc"),
 *    where("created_at", ">",  now),
 *  )
 * ).execute('http:...').then(response => {
 *   for (const result of response.data) {
 *     if (isNamed(result, 'latest-games')) {
 *       result.result.name // result is of type `{ name: string }`
 *     } else {
 *       result.result.created_at  // result is of type `{ created_at: number }`
 *     }
 *   }
 * });
 * ```
 * @param builder
 * @param name
 */
function isNamed(builder, name) {
    if (builder.name === name) {
        return true;
    }
    return false;
}
exports.isNamed = isNamed;
//# sourceMappingURL=index.js.map